/*
    split: a mini-app to split a string based on a delimiter.
        Useful for delimited arguments for shell scripts
    Copyright (C) 2006, Aaron Griffin <aaronmgriffin@gmail.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <string.h>

int main(int argc, char **argv)
{
    char* tok = NULL;

    if(argc != 3)
        printf("usage: %s <string> <delimiter>",argv[0]);

    tok = strtok(argv[1],argv[2]);
    while(tok)
    {
        printf("%s",tok);
        tok = strtok(NULL,argv[2]);
        if(tok) printf(" ");
    }
    printf("\n");

    return 0;
}

